# sage_setup: distribution = sagemath-graphs
r"""
Graphs from the World Map

The methods defined here appear in :mod:`sage.graphs.graph_generators`.
"""

# ****************************************************************************
#           Copyright (C) 2006 Robert L. Miller <rlmillster@gmail.com>
#                              and Emily A. Kirkman
#           Copyright (C) 2009 Michael C. Yurko <myurko@gmail.com>
#
# Distributed  under  the  terms  of  the  GNU  General  Public  License (GPL)
#                         http://www.gnu.org/licenses/
# ****************************************************************************

# import from Sage library
from sage.graphs.graph import Graph


def AfricaMap(continental=False, year=2018):
    """
    Return African states as a graph of common border.

    "African state" here is defined as an independent state having the capital
    city in Africa. The graph has an edge between those countries that have
    common *land* border.

    INPUT:

    - ``continental`` -- boolean (default: ``False``); whether to only return
      states in the continental Africa or all African states

    - ``year`` -- integer (default: 2018); reserved for future use

    EXAMPLES::

        sage: Africa = graphs.AfricaMap(); Africa
        Africa Map: Graph on 54 vertices
        sage: sorted(Africa.neighbors('Libya'))
        ['Algeria', 'Chad', 'Egypt', 'Niger', 'Sudan', 'Tunisia']

        sage: cont_Africa = graphs.AfricaMap(continental=True)
        sage: cont_Africa.order()
        48
        sage: 'Madagaskar' in cont_Africa
        False

    TESTS::

        sage: Africa.plot()                     # long time                             # needs sage.plot
        Graphics object consisting of 159 graphics primitives
    """
    if year != 2018:
        raise ValueError("currently only year 2018 is implemented")

    common_border = {
     'Algeria': ['Libya', 'Mali', 'Mauritania', 'Morocco', 'Niger', 'Tunisia'],
     'Angola': ['Namibia', 'Zambia'],
     'Benin': ['Burkina Faso', 'Niger', 'Nigeria', 'Togo'],
     'Botswana': ['Namibia', 'South Africa', 'Zimbabwe'],
     'Burkina Faso': ['Ghana', 'Ivory Coast', 'Mali', 'Niger', 'Togo'],
     'Cameroon': ['Central Africa', 'Chad', 'Equatorial Guinea', 'Gabon',
                  'Nigeria'],
     'Central Africa': ['Chad', 'South Sudan', 'Sudan'],
     'Chad': ['Libya', 'Niger', 'Nigeria', 'Sudan'],
     'Republic of the Congo': ['Gabon', 'Cameroon', 'Central Africa', 'Angola',
                               'Democratic Republic of the Congo'],
     'Democratic Republic of the Congo': ['Zambia', 'South Sudan', 'Tanzania',
                                          'Burundi', 'Rwanda', 'Uganda',
                                          'Central Africa', 'Angola'],
     'Djibouti': ['Eritrea', 'Ethiopia', 'Somalia'],
     'Ethiopia': ['Eritrea', 'Kenya', 'Somalia', 'South Sudan', 'Sudan'],
     'Gabon': ['Equatorial Guinea'],
     'Ghana': ['Ivory Coast', 'Togo'],
     'Guinea': ['Guinea-Bissau', 'Ivory Coast', 'Liberia', 'Sierra Leone'],
     'Kenya': ['Somalia', 'South Sudan', 'Tanzania', 'Uganda'],
     'Liberia': ['Ivory Coast', 'Sierra Leone'],
     'Libya': ['Egypt', 'Niger', 'Sudan', 'Tunisia'],
     'Mali': ['Guinea', 'Ivory Coast', 'Mauritania', 'Niger', 'Senegal'],
     'Mozambique': ['Malawi', 'South Africa', 'Swaziland', 'Zimbabwe'],
     'Niger': ['Nigeria'],
     'Rwanda': ['Burundi', 'Tanzania', 'Uganda'],
     'Senegal': ['Guinea', 'Guinea-Bissau', 'Mauritania', 'Gambia'],
     'South Africa': ['Lesotho', 'Namibia', 'Swaziland', 'Zimbabwe'],
     'South Sudan': ['Uganda', 'Sudan', 'Democratic Republic of the Congo'],
     'Sudan': ['Egypt', 'Eritrea'],
     'Tanzania': ['Burundi', 'Malawi', 'Mozambique', 'Uganda', 'Zambia'],
     'Zambia': ['Malawi', 'Mozambique', 'Namibia', 'Zimbabwe']
     }

    no_land_border = ['Cape Verde', 'Seychelles', 'Mauritius',
                      'São Tomé and Príncipe', 'Madagascar', 'Comoros']

    G = Graph(common_border, format='dict_of_lists')

    if continental:
        G = G.subgraph(G.connected_component_containing_vertex('Central Africa', sort=False))
        G.name(new="Continental Africa Map")
    else:
        G.add_vertices(no_land_border)
        G.name(new="Africa Map")

    return G


def EuropeMap(continental=False, year=2018):
    """
    Return European states as a graph of common border.

    "European state" here is defined as an independent state having the capital
    city in Europe. The graph has an edge between those countries that have
    common *land* border.

    INPUT:

    - ``continental`` -- boolean (default: ``False``); whether to only return
      states in the continental Europe or all European states

    - ``year`` -- integer (default: 2018); reserved for future use

    EXAMPLES::

        sage: Europe = graphs.EuropeMap(); Europe
        Europe Map: Graph on 44 vertices
        sage: Europe.neighbors('Ireland')
        ['United Kingdom']

        sage: cont_Europe = graphs.EuropeMap(continental=True)
        sage: cont_Europe.order()
        40
        sage: 'Iceland' in cont_Europe
        False
    """
    if year != 2018:
        raise ValueError("currently only year 2018 is implemented")

    common_border = {
     'Austria': ['Czech Republic', 'Germany', 'Liechtenstein', 'Slovenia',
                 'Switzerland'],
     'Belarus': ['Latvia', 'Lithuania', 'Poland', 'Russia', 'Ukraine'],
     'Belgium': ['France', 'Germany', 'Luxembourg', 'Netherlands'],
     'Croatia': ['Bosnia and Herzegovina', 'Hungary', 'Montenegro', 'Serbia',
                 'Slovenia'],
     'France': ['Andorra', 'Germany', 'Italy', 'Luxembourg', 'Monaco',
                'Switzerland'],
     'Germany': ['Czech Republic', 'Denmark', 'Luxembourg', 'Netherlands',
                 'Switzerland'],
     'Greece': ['Albania', 'Bulgaria', 'Macedonia'],
     'Hungary': ['Austria', 'Romania', 'Serbia', 'Slovakia', 'Slovenia',
                 'Ukraine'],
     'Ireland': ['United Kingdom'],
     'Italy': ['Austria', 'San Marino', 'Slovenia', 'Switzerland',
               'Vatican City'],
     'Latvia': ['Estonia', 'Lithuania', 'Russia'],
     'Macedonia': ['Albania', 'Bulgaria', 'Serbia'],
     'Montenegro': ['Albania', 'Bosnia and Herzegovina', 'Serbia'],
     'Norway': ['Finland', 'Russia', 'Sweden'],
     'Poland': ['Czech Republic', 'Germany', 'Lithuania', 'Russia', 'Slovakia',
                'Ukraine'],
     'Romania': ['Bulgaria', 'Moldova', 'Serbia', 'Ukraine'],
     'Russia': ['Estonia', 'Finland', 'Lithuania', 'Ukraine'],
     'Serbia': ['Bosnia and Herzegovina', 'Bulgaria'],
     'Slovakia': ['Austria', 'Czech Republic', 'Ukraine'],
     'Spain': ['Andorra', 'France', 'Portugal'],
     'Sweden': ['Finland'],
     'Switzerland': ['Liechtenstein'],
     'Ukraine': ['Moldova']
    }
    no_land_border = ['Iceland', 'Malta']

    G = Graph(common_border, format='dict_of_lists')

    if continental:
        G = G.subgraph(G.connected_component_containing_vertex('Austria', sort=False))
        G.name(new="Continental Europe Map")
    else:
        G.add_vertices(no_land_border)
        G.name(new="Europe Map")

    return G


def USAMap(continental=False):
    """
    Return states of USA as a graph of common border.

    The graph has an edge between those states that have common *land* border
    line or point. Hence for example Colorado and Arizona are marked as
    neighbors, but Michigan and Minnesota are not.

    INPUT:

    - ``continental`` -- boolean (default: ``False``); whether to exclude Alaska
      and Hawaii

    EXAMPLES:

    How many states are neighbor's neighbor for Pennsylvania::

        sage: USA = graphs.USAMap()
        sage: distance = USA.shortest_path_lengths('Pennsylvania')
        sage: len([n2 for n2, d in distance.items() if d == 2])
        7

    Diameter for continental USA::

        sage: USAcont = graphs.USAMap(continental=True)
        sage: USAcont.diameter()
        11
    """
    states = {
        "Alabama": ["Florida", "Georgia", "Mississippi", "Tennessee"],
        "Arizona": ["California", "Colorado", "Nevada", "New Mexico", "Utah"],
        "Arkansas": ["Louisiana", "Mississippi", "Missouri", "Oklahoma",
                     "Tennessee", "Texas"],
        "California": ["Arizona", "Nevada", "Oregon"],
        "Colorado": ["Arizona", "Kansas", "Nebraska", "New Mexico", "Oklahoma",
                     "Utah", "Wyoming"],
        "Connecticut": ["Massachusetts", "New York", "Rhode Island"],
        "Delaware": ["Maryland", "New Jersey", "Pennsylvania"],
        "Florida": ["Alabama", "Georgia"],
        "Georgia": ["Alabama", "Florida", "North Carolina", "South Carolina",
                    "Tennessee"],
        "Idaho": ["Montana", "Nevada", "Oregon", "Utah", "Washington",
                  "Wyoming"],
        "Illinois": ["Indiana", "Iowa", "Michigan", "Kentucky", "Missouri",
                     "Wisconsin"],
        "Indiana": ["Illinois", "Kentucky", "Michigan", "Ohio"],
        "Iowa": ["Illinois", "Minnesota", "Missouri", "Nebraska",
                 "South Dakota", "Wisconsin"],
        "Kansas": ["Colorado", "Missouri", "Nebraska", "Oklahoma"],
        "Kentucky": ["Illinois", "Indiana", "Missouri", "Ohio", "Tennessee",
                     "Virginia", "West Virginia"],
        "Louisiana": ["Arkansas", "Mississippi", "Texas"],
        "Maine": ["New Hampshire"],
        "Maryland": ["Delaware", "Pennsylvania", "Virginia", "West Virginia"],
        "Massachusetts": ["Connecticut", "New Hampshire", "New York",
                          "Rhode Island", "Vermont"],
        "Michigan": ["Illinois", "Indiana", "Ohio", "Wisconsin"],
        "Minnesota": ["Iowa", "North Dakota", "South Dakota", "Wisconsin"],
        "Mississippi": ["Alabama", "Arkansas", "Louisiana", "Tennessee"],
        "Missouri": ["Arkansas", "Illinois", "Iowa", "Kansas", "Kentucky",
                     "Nebraska", "Oklahoma", "Tennessee"],
        "Montana": ["Idaho", "North Dakota", "South Dakota", "Wyoming"],
        "Nebraska": ["Colorado", "Iowa", "Kansas", "Missouri", "South Dakota",
                     "Wyoming"],
        "Nevada": ["Arizona", "California", "Idaho", "Oregon", "Utah"],
        "New Hampshire": ["Maine", "Massachusetts", "Vermont"],
        "New Jersey": ["Delaware", "New York", "Pennsylvania"],
        "New Mexico": ["Arizona", "Colorado", "Oklahoma", "Texas", "Utah"],
        "New York": ["Connecticut", "Massachusetts", "New Jersey",
                     "Pennsylvania", "Vermont"],
        "North Carolina": ["Georgia", "South Carolina", "Tennessee",
                           "Virginia"],
        "North Dakota": ["Minnesota", "Montana", "South Dakota"],
        "Ohio": ["Indiana", "Kentucky", "Michigan", "Pennsylvania",
                 "West Virginia"],
        "Oklahoma": ["Arkansas", "Colorado", "Kansas", "Missouri",
                     "New Mexico", "Texas"],
        "Oregon": ["California", "Idaho", "Nevada", "Washington"],
        "Pennsylvania": ["Delaware", "Maryland", "New Jersey", "New York",
                         "Ohio", "West Virginia"],
        "Rhode Island": ["Connecticut", "Massachusetts"],
        "South Carolina": ["Georgia", "North Carolina"],
        "South Dakota": ["Iowa", "Minnesota", "Montana", "Nebraska",
                         "North Dakota", "Wyoming"],
        "Tennessee": ["Alabama", "Arkansas", "Georgia", "Kentucky",
                      "Mississippi", "Missouri", "North Carolina", "Virginia"],
        "Texas": ["Arkansas", "Louisiana", "New Mexico", "Oklahoma"],
        "Utah": ["Arizona", "Colorado", "Idaho", "Nevada", "New Mexico",
                 "Wyoming"],
        "Vermont": ["Massachusetts", "New Hampshire", "New York"],
        "Virginia": ["Kentucky", "Maryland", "North Carolina", "Tennessee",
                     "West Virginia"],
        "Washington": ["Idaho", "Oregon"],
        "West Virginia": ["Kentucky", "Maryland", "Ohio", "Pennsylvania",
                          "Virginia"],
        "Wisconsin": ["Illinois", "Iowa", "Michigan", "Minnesota"],
        "Wyoming": ["Colorado", "Idaho", "Montana", "Nebraska", "South Dakota",
                    "Utah"]
    }
    if continental:
        name = "Continental USA Map"
    else:
        states['Alaska'] = []
        states['Hawaii'] = []
        name = "USA Map"

    return Graph(states, format='dict_of_lists', name=name)


def WorldMap():
    """
    Return the Graph of all the countries, in which two countries are adjacent
    in the graph if they have a common boundary.

    This graph has been built from the data available
    in The CIA World Factbook [CIA]_ (2009-08-21).

    The returned graph ``G`` has a member ``G.gps_coordinates`` equal to a
    dictionary containing the GPS coordinates of each country's capital city.

    EXAMPLES::

        sage: g = graphs.WorldMap()
        sage: g.has_edge("France", "Italy")
        True
        sage: g.gps_coordinates["Bolivia"]
        [[17, 'S'], [65, 'W']]
        sage: g.connected_component_containing_vertex('Ireland', sort=True)
        ['Ireland', 'United Kingdom']

    TESTS:

    :issue:`24488`::

        sage: 'Iceland' in graphs.WorldMap()
        True
    """
    edges = [
        ('Afghanistan', 'China'), ('Afghanistan', 'Iran'),
        ('Afghanistan', 'Uzbekistan'), ('Albania', 'Greece'),
        ('Albania', 'Kosovo'), ('Albania', 'Macedonia'),
        ('Albania', 'Montenegro'), ('Algeria', 'Morocco'),
        ('Algeria', 'Tunisia'), ('Andorra', 'Spain'),
        ('Angola', 'Democratic Republic of the Congo'), ('Angola', 'Namibia'),
        ('Angola', 'Zambia'), ('Argentina', 'Bolivia'), ('Argentina', 'Brazil'),
        ('Argentina', 'Chile'), ('Argentina', 'Paraguay'),
        ('Argentina', 'Uruguay'), ('Armenia', 'Georgia'), ('Armenia', 'Iran'),
        ('Austria', 'Germany'), ('Azerbaijan', 'Armenia'),
        ('Azerbaijan', 'Georgia'), ('Azerbaijan', 'Iran'),
        ('Azerbaijan', 'Russia'), ('Azerbaijan', 'Turkey'),
        ('Bangladesh', 'Burma'), ('Belgium', 'Germany'),
        ('Belgium', 'Netherlands'), ('Belize', 'Mexico'),
        ('Benin', 'Burkina Faso'), ('Benin', 'Niger'), ('Benin', 'Nigeria'),
        ('Benin', 'Togo'), ('Bolivia', 'Brazil'), ('Bolivia', 'Chile'),
        ('Bolivia', 'Paraguay'), ('Bolivia', 'Peru'),
        ('Bosnia and Herzegovina', 'Croatia'),
        ('Bosnia and Herzegovina', 'Montenegro'),
        ('Bosnia and Herzegovina', 'Serbia'), ('Brazil', 'Colombia'),
        ('Brazil', 'Guyana'), ('Brazil', 'Suriname'), ('Brazil', 'Venezuela'),
        ('Bulgaria', 'Greece'), ('Bulgaria', 'Macedonia'),
        ('Bulgaria', 'Romania'), ('Bulgaria', 'Serbia'),
        ('Burkina Faso', 'Mali'), ('Burkina Faso', 'Niger'),
        ('Burkina Faso', 'Togo'),
        ('Burundi', 'Democratic Republic of the Congo'), ('Cambodia', 'Laos'),
        ('Cambodia', 'Thailand'), ('Cambodia', 'Vietnam'),
        ('Cameroon', 'Central African Republic'), ('Cameroon', 'Chad'),
        ('Cameroon', 'Equatorial Guinea'), ('Cameroon', 'Nigeria'),
        ('Cameroon', 'Republic of the Congo'), ('Canada', 'United States'),
        ('Central African Republic', 'Chad'),
        ('Central African Republic', 'Democratic Republic of the Congo'),
        ('Central African Republic', 'Sudan'), ('Chad', 'Niger'),
        ('Chad', 'Nigeria'), ('Chad', 'Sudan'), ('China', 'Bhutan'),
        ('China', 'Burma'), ('China', 'Hong Kong'), ('China', 'Kazakhstan'),
        ('China', 'Kyrgyzstan'), ('China', 'Mongolia'), ('China', 'Nepal'),
        ('China', 'North Korea'), ('China', 'Russia'), ('China', 'Vietnam'),
        ('Colombia', 'Venezuela'), ('Costa Rica', 'Nicaragua'),
        ("Cote d'Ivoire", 'Burkina Faso'), ("Cote d'Ivoire", 'Guinea'),
        ("Cote d'Ivoire", 'Mali'), ('Cyprus', 'Akrotiri'),
        ('Cyprus', 'Dhekelia'), ('Czech Republic', 'Austria'),
        ('Czech Republic', 'Germany'), ('Czech Republic', 'Poland'),
        ('Democratic Republic of the Congo', 'Zambia'), ('Denmark', 'Germany'),
        ('Djibouti', 'Eritrea'), ('Dominican Republic', 'Haiti'),
        ('Ecuador', 'Colombia'), ('El Salvador', 'Honduras'),
        ('Ethiopia', 'Djibouti'), ('Ethiopia', 'Eritrea'),
        ('Ethiopia', 'Kenya'), ('Ethiopia', 'Somalia'), ('Ethiopia', 'Sudan'),
        ('Finland', 'Russia'), ('Finland', 'Sweden'), ('France', 'Andorra'),
        ('France', 'Belgium'), ('France', 'Brazil'), ('France', 'Germany'),
        ('France', 'Italy'), ('France', 'Luxembourg'), ('France', 'Spain'),
        ('France', 'Suriname'), ('France', 'Switzerland'),
        ('Gabon', 'Cameroon'), ('Gabon', 'Equatorial Guinea'),
        ('Gabon', 'Republic of the Congo'), ('Gaza Strip', 'Egypt'),
        ('Gaza Strip', 'Israel'), ('Ghana', 'Burkina Faso'),
        ('Ghana', "Cote d'Ivoire"), ('Ghana', 'Togo'), ('Gibraltar', 'Spain'),
        ('Guatemala', 'Belize'), ('Guatemala', 'El Salvador'),
        ('Guatemala', 'Honduras'), ('Guatemala', 'Mexico'),
        ('Guinea', 'Sierra Leone'), ('Guinea-Bissau', 'Guinea'),
        ('Guinea-Bissau', 'Senegal'), ('Honduras', 'Nicaragua'),
        ('Hungary', 'Austria'), ('Hungary', 'Croatia'), ('Hungary', 'Serbia'),
        ('India', 'Bangladesh'), ('India', 'Bhutan'), ('India', 'Burma'),
        ('India', 'China'), ('India', 'Nepal'),
        ('Indonesia', 'Papua New Guinea'), ('Iran', 'Iraq'),
        ('Ireland', 'United Kingdom'), ('Israel', 'Egypt'),
        ('Italy', 'Austria'), ('Jordan', 'Iraq'), ('Jordan', 'Israel'),
        ('Jordan', 'Syria'), ('Jordan', 'West Bank'),
        ('Kazakhstan', 'Kyrgyzstan'), ('Kenya', 'Somalia'), ('Kenya', 'Sudan'),
        ('Kenya', 'Uganda'), ('Kosovo', 'Macedonia'), ('Kosovo', 'Serbia'),
        ('Kuwait', 'Iraq'), ('Laos', 'Burma'), ('Laos', 'China'),
        ('Laos', 'Thailand'), ('Laos', 'Vietnam'), ('Latvia', 'Belarus'),
        ('Latvia', 'Estonia'), ('Lebanon', 'Israel'),
        ('Lesotho', 'South Africa'), ('Liberia', "Cote d'Ivoire"),
        ('Liberia', 'Guinea'), ('Liberia', 'Sierra Leone'),
        ('Libya', 'Algeria'), ('Libya', 'Chad'), ('Libya', 'Egypt'),
        ('Libya', 'Niger'), ('Libya', 'Sudan'), ('Libya', 'Tunisia'),
        ('Liechtenstein', 'Austria'), ('Liechtenstein', 'Switzerland'),
        ('Lithuania', 'Belarus'), ('Lithuania', 'Latvia'),
        ('Lithuania', 'Poland'), ('Lithuania', 'Russia'),
        ('Luxembourg', 'Belgium'), ('Luxembourg', 'Germany'),
        ('Macau', 'China'), ('Macedonia', 'Greece'), ('Macedonia', 'Serbia'),
        ('Malaysia', 'Brunei'), ('Malaysia', 'Indonesia'),
        ('Malaysia', 'Thailand'), ('Mali', 'Algeria'), ('Mali', 'Guinea'),
        ('Mali', 'Niger'), ('Mali', 'Senegal'), ('Mauritania', 'Algeria'),
        ('Mauritania', 'Mali'), ('Mauritania', 'Senegal'),
        ('Mauritania', 'Western Sahara'), ('Monaco', 'France'),
        ('Montenegro', 'Croatia'), ('Montenegro', 'Kosovo'),
        ('Montenegro', 'Serbia'), ('Morocco', 'Spain'),
        ('Mozambique', 'Malawi'), ('Mozambique', 'Zambia'),
        ('Mozambique', 'Zimbabwe'), ('Namibia', 'Botswana'),
        ('Namibia', 'Zambia'), ('Netherlands', 'Germany'), ('Niger', 'Algeria'),
        ('Niger', 'Nigeria'), ('Norway', 'Finland'), ('Norway', 'Russia'),
        ('Norway', 'Sweden'), ('Oman', 'United Arab Emirates'),
        ('Oman', 'Yemen'), ('Pakistan', 'Afghanistan'), ('Pakistan', 'China'),
        ('Pakistan', 'India'), ('Pakistan', 'Iran'), ('Panama', 'Colombia'),
        ('Panama', 'Costa Rica'), ('Paraguay', 'Brazil'), ('Peru', 'Brazil'),
        ('Peru', 'Chile'), ('Peru', 'Colombia'), ('Peru', 'Ecuador'),
        ('Poland', 'Belarus'), ('Poland', 'Germany'), ('Portugal', 'Spain'),
        ('Republic of the Congo', 'Angola'),
        ('Republic of the Congo', 'Central African Republic'),
        ('Republic of the Congo', 'Democratic Republic of the Congo'),
        ('Romania', 'Hungary'), ('Romania', 'Moldova'), ('Romania', 'Serbia'),
        ('Russia', 'Belarus'), ('Russia', 'Estonia'), ('Russia', 'Georgia'),
        ('Russia', 'Kazakhstan'), ('Russia', 'Latvia'), ('Russia', 'Mongolia'),
        ('Russia', 'North Korea'), ('Russia', 'Poland'), ('Rwanda', 'Burundi'),
        ('Rwanda', 'Democratic Republic of the Congo'), ('Rwanda', 'Uganda'),
        ('Saint Martin', 'Netherlands Antilles'), ('San Marino', 'Italy'),
        ('Saudi Arabia', 'Iraq'), ('Saudi Arabia', 'Jordan'),
        ('Saudi Arabia', 'Kuwait'), ('Saudi Arabia', 'Oman'),
        ('Saudi Arabia', 'Qatar'), ('Saudi Arabia', 'United Arab Emirates'),
        ('Saudi Arabia', 'Yemen'), ('Senegal', 'Guinea'), ('Serbia', 'Croatia'),
        ('Slovakia', 'Austria'), ('Slovakia', 'Czech Republic'),
        ('Slovakia', 'Hungary'), ('Slovakia', 'Poland'),
        ('Slovakia', 'Ukraine'), ('Slovenia', 'Austria'),
        ('Slovenia', 'Croatia'), ('Slovenia', 'Hungary'), ('Slovenia', 'Italy'),
        ('Somalia', 'Djibouti'), ('South Africa', 'Botswana'),
        ('South Africa', 'Mozambique'), ('South Africa', 'Namibia'),
        ('South Africa', 'Zimbabwe'), ('South Korea', 'North Korea'),
        ('Sudan', 'Democratic Republic of the Congo'), ('Sudan', 'Egypt'),
        ('Sudan', 'Eritrea'), ('Suriname', 'Guyana'),
        ('Swaziland', 'Mozambique'), ('Swaziland', 'South Africa'),
        ('Switzerland', 'Austria'), ('Switzerland', 'Germany'),
        ('Switzerland', 'Italy'), ('Syria', 'Iraq'), ('Syria', 'Israel'),
        ('Syria', 'Lebanon'), ('Tajikistan', 'Afghanistan'),
        ('Tajikistan', 'China'), ('Tajikistan', 'Kyrgyzstan'),
        ('Tajikistan', 'Uzbekistan'), ('Tanzania', 'Burundi'),
        ('Tanzania', 'Democratic Republic of the Congo'), ('Tanzania', 'Kenya'),
        ('Tanzania', 'Malawi'), ('Tanzania', 'Mozambique'),
        ('Tanzania', 'Rwanda'), ('Tanzania', 'Uganda'), ('Tanzania', 'Zambia'),
        ('Thailand', 'Burma'), ('The Gambia', 'Senegal'),
        ('Timor-Leste', 'Indonesia'), ('Turkey', 'Armenia'),
        ('Turkey', 'Bulgaria'), ('Turkey', 'Georgia'), ('Turkey', 'Greece'),
        ('Turkey', 'Iran'), ('Turkey', 'Iraq'), ('Turkey', 'Syria'),
        ('Turkmenistan', 'Afghanistan'), ('Turkmenistan', 'Iran'),
        ('Turkmenistan', 'Kazakhstan'), ('Turkmenistan', 'Uzbekistan'),
        ('Uganda', 'Democratic Republic of the Congo'), ('Uganda', 'Sudan'),
        ('Ukraine', 'Belarus'), ('Ukraine', 'Hungary'), ('Ukraine', 'Moldova'),
        ('Ukraine', 'Poland'), ('Ukraine', 'Romania'), ('Ukraine', 'Russia'),
        ('United States', 'Mexico'), ('Uruguay', 'Brazil'),
        ('Uzbekistan', 'Kazakhstan'), ('Uzbekistan', 'Kyrgyzstan'),
        ('Vatican City', 'Italy'), ('Venezuela', 'Guyana'),
        ('West Bank', 'Israel'), ('Western Sahara', 'Algeria'),
        ('Western Sahara', 'Morocco'), ('Zambia', 'Malawi'),
        ('Zambia', 'Zimbabwe'), ('Zimbabwe', 'Botswana')
        ]
    gps_coordinates = {
        "Cote d'Ivoire": [[8, 'N'], [5, 'W']],
        'Afghanistan': [[33, 'N'], [65, 'E']],
        'Akrotiri': [[34, 'N'], [32, 'E']],
        'Albania': [[41, 'N'], [20, 'E']],
        'Algeria': [[28, 'N'], [3, 'E']],
        'American Samoa': [[14, 'S'], [170, 'W']],
        'Andorra': [[42, 'N'], [1, 'E']],
        'Angola': [[12, 'S'], [18, 'E']],
        'Anguilla': [[18, 'N'], [63, 'W']],
        'Antarctica': [[90, 'S'], [0, 'E']],
        'Antigua and Barbuda': [[17, 'N'], [61, 'W']],
        'Argentina': [[34, 'S'], [64, 'W']],
        'Armenia': [[40, 'N'], [45, 'E']],
        'Aruba': [[12, 'N'], [69, 'W']],
        'Ashmore and Cartier Islands': [[12, 'S'], [123, 'E']],
        'Australia': [[27, 'S'], [133, 'E']],
        'Austria': [[47, 'N'], [13, 'E']],
        'Azerbaijan': [[40, 'N'], [47, 'E']],
        'Bahamas, The': [[24, 'N'], [76, 'W']],
        'Bahrain': [[26, 'N'], [50, 'E']],
        'Bangladesh': [[24, 'N'], [90, 'E']],
        'Barbados': [[13, 'N'], [59, 'W']],
        'Belarus': [[53, 'N'], [28, 'E']],
        'Belgium': [[50, 'N'], [4, 'E']],
        'Belize': [[17, 'N'], [88, 'W']],
        'Benin': [[9, 'N'], [2, 'E']],
        'Bermuda': [[32, 'N'], [64, 'W']],
        'Bhutan': [[27, 'N'], [90, 'E']],
        'Bolivia': [[17, 'S'], [65, 'W']],
        'Bosnia and Herzegovina': [[44, 'N'], [18, 'E']],
        'Botswana': [[22, 'S'], [24, 'E']],
        'Bouvet Island': [[54, 'S'], [3, 'E']],
        'Brazil': [[10, 'S'], [55, 'W']],
        'British Indian Ocean Territory': [[6, 'S'], [71, 'E']],
        'British Virgin Islands': [[18, 'N'], [64, 'W']],
        'Brunei': [[4, 'N'], [114, 'E']],
        'Bulgaria': [[43, 'N'], [25, 'E']],
        'Burkina Faso': [[13, 'N'], [2, 'W']],
        'Burma': [[22, 'N'], [98, 'E']],
        'Burundi': [[3, 'S'], [30, 'E']],
        'Cambodia': [[13, 'N'], [105, 'E']],
        'Cameroon': [[6, 'N'], [12, 'E']],
        'Canada': [[60, 'N'], [95, 'W']],
        'Cape Verde': [[16, 'N'], [24, 'W']],
        'Cayman Islands': [[19, 'N'], [80, 'W']],
        'Central African Republic': [[7, 'N'], [21, 'E']],
        'Chad': [[15, 'N'], [19, 'E']],
        'Chile': [[30, 'S'], [71, 'W']],
        'China': [[35, 'N'], [105, 'E']],
        'Christmas Island': [[10, 'S'], [105, 'E']],
        'Clipperton Island': [[10, 'N'], [109, 'W']],
        'Cocos (Keeling) Islands': [[12, 'S'], [96, 'E']],
        'Colombia': [[4, 'N'], [72, 'W']],
        'Comoros': [[12, 'S'], [44, 'E']],
        'Cook Islands': [[21, 'S'], [159, 'W']],
        'Coral Sea Islands': [[18, 'S'], [152, 'E']],
        'Costa Rica': [[10, 'N'], [84, 'W']],
        'Croatia': [[45, 'N'], [15, 'E']],
        'Cuba': [[21, 'N'], [80, 'W']],
        'Cyprus': [[35, 'N'], [33, 'E']],
        'Czech Republic': [[49, 'N'], [15, 'E']],
        'Democratic Republic of the Congo': [[0, 'N'], [25, 'E']],
        'Denmark': [[56, 'N'], [10, 'E']],
        'Dhekelia': [[34, 'N'], [33, 'E']],
        'Djibouti': [[11, 'N'], [43, 'E']],
        'Dominica': [[15, 'N'], [61, 'W']],
        'Dominican Republic': [[19, 'N'], [70, 'W']],
        'Ecuador': [[2, 'S'], [77, 'W']],
        'Egypt': [[27, 'N'], [30, 'E']],
        'El Salvador': [[13, 'N'], [88, 'W']],
        'Equatorial Guinea': [[2, 'N'], [10, 'E']],
        'Eritrea': [[15, 'N'], [39, 'E']],
        'Estonia': [[59, 'N'], [26, 'E']],
        'Ethiopia': [[8, 'N'], [38, 'E']],
        'Falkland Islands (Islas Malvinas)': [[51, 'S'], [59, 'W']],
        'Faroe Islands': [[62, 'N'], [7, 'W']],
        'Fiji': [[18, 'S'], [175, 'E']],
        'Finland': [[64, 'N'], [26, 'E']],
        'France': [[46, 'N'], [2, 'E']],
        'French Polynesia': [[15, 'S'], [140, 'W']],
        'Gabon': [[1, 'S'], [11, 'E']],
        'Gaza Strip': [[31, 'N'], [34, 'E']],
        'Georgia': [[42, 'N'], [43, 'E']],
        'Germany': [[51, 'N'], [9, 'E']],
        'Ghana': [[8, 'N'], [2, 'W']],
        'Gibraltar': [[36, 'N'], [5, 'W']],
        'Greece': [[39, 'N'], [22, 'E']],
        'Greenland': [[72, 'N'], [40, 'W']],
        'Grenada': [[12, 'N'], [61, 'W']],
        'Guam': [[13, 'N'], [144, 'E']],
        'Guatemala': [[15, 'N'], [90, 'W']],
        'Guernsey': [[49, 'N'], [2, 'W']],
        'Guinea': [[11, 'N'], [10, 'W']],
        'Guinea-Bissau': [[12, 'N'], [15, 'W']],
        'Guyana': [[5, 'N'], [59, 'W']],
        'Haiti': [[19, 'N'], [72, 'W']],
        'Heard Island and McDonald Islands': [[53, 'S'], [72, 'E']],
        'Honduras': [[15, 'N'], [86, 'W']],
        'Hong Kong': [[22, 'N'], [114, 'E']],
        'Hungary': [[47, 'N'], [20, 'E']],
        'Iceland': [[65, 'N'], [18, 'W']],
        'India': [[20, 'N'], [77, 'E']],
        'Indonesia': [[5, 'S'], [120, 'E']],
        'Iran': [[32, 'N'], [53, 'E']],
        'Iraq': [[33, 'N'], [44, 'E']],
        'Ireland': [[53, 'N'], [8, 'W']],
        'Isle of Man': [[54, 'N'], [4, 'W']],
        'Israel': [[31, 'N'], [34, 'E']],
        'Italy': [[42, 'N'], [12, 'E']],
        'Jamaica': [[18, 'N'], [77, 'W']],
        'Jan Mayen': [[71, 'N'], [8, 'W']],
        'Japan': [[36, 'N'], [138, 'E']],
        'Jersey': [[49, 'N'], [2, 'W']],
        'Jordan': [[31, 'N'], [36, 'E']],
        'Kazakhstan': [[48, 'N'], [68, 'E']],
        'Kenya': [[1, 'N'], [38, 'E']],
        'Kiribati': [[1, 'N'], [173, 'E']],
        'Kosovo': [[42, 'N'], [21, 'E']],
        'Kuwait': [[29, 'N'], [45, 'E']],
        'Kyrgyzstan': [[41, 'N'], [75, 'E']],
        'Laos': [[18, 'N'], [105, 'E']],
        'Latvia': [[57, 'N'], [25, 'E']],
        'Lebanon': [[33, 'N'], [35, 'E']],
        'Lesotho': [[29, 'S'], [28, 'E']],
        'Liberia': [[6, 'N'], [9, 'W']],
        'Libya': [[25, 'N'], [17, 'E']],
        'Liechtenstein': [[47, 'N'], [9, 'E']],
        'Lithuania': [[56, 'N'], [24, 'E']],
        'Luxembourg': [[49, 'N'], [6, 'E']],
        'Macau': [[22, 'N'], [113, 'E']],
        'Macedonia': [[41, 'N'], [22, 'E']],
        'Madagascar': [[20, 'S'], [47, 'E']],
        'Malawi': [[13, 'S'], [34, 'E']],
        'Malaysia': [[2, 'N'], [112, 'E']],
        'Maldives': [[3, 'N'], [73, 'E']],
        'Mali': [[17, 'N'], [4, 'W']],
        'Malta': [[35, 'N'], [14, 'E']],
        'Marshall Islands': [[9, 'N'], [168, 'E']],
        'Mauritania': [[20, 'N'], [12, 'W']],
        'Mauritius': [[20, 'S'], [57, 'E']],
        'Mayotte': [[12, 'S'], [45, 'E']],
        'Mexico': [[23, 'N'], [102, 'W']],
        'Micronesia, Federated States of': [[6, 'N'], [158, 'E']],
        'Moldova': [[47, 'N'], [29, 'E']],
        'Monaco': [[43, 'N'], [7, 'E']],
        'Mongolia': [[46, 'N'], [105, 'E']],
        'Montenegro': [[42, 'N'], [19, 'E']],
        'Montserrat': [[16, 'N'], [62, 'W']],
        'Morocco': [[32, 'N'], [5, 'W']],
        'Mozambique': [[18, 'S'], [35, 'E']],
        'Namibia': [[22, 'S'], [17, 'E']],
        'Nauru': [[0, 'S'], [166, 'E']],
        'Navassa Island': [[18, 'N'], [75, 'W']],
        'Nepal': [[28, 'N'], [84, 'E']],
        'Netherlands Antilles': [[12, 'N'], [69, 'W']],
        'Netherlands': [[52, 'N'], [5, 'E']],
        'New Caledonia': [[21, 'S'], [165, 'E']],
        'New Zealand': [[41, 'S'], [174, 'E']],
        'Nicaragua': [[13, 'N'], [85, 'W']],
        'Niger': [[16, 'N'], [8, 'E']],
        'Nigeria': [[10, 'N'], [8, 'E']],
        'Niue': [[19, 'S'], [169, 'W']],
        'Norfolk Island': [[29, 'S'], [167, 'E']],
        'North Korea': [[40, 'N'], [127, 'E']],
        'Northern Mariana Islands': [[15, 'N'], [145, 'E']],
        'Norway': [[62, 'N'], [10, 'E']],
        'Oman': [[21, 'N'], [57, 'E']],
        'Pakistan': [[30, 'N'], [70, 'E']],
        'Palau': [[7, 'N'], [134, 'E']],
        'Panama': [[9, 'N'], [80, 'W']],
        'Papua New Guinea': [[6, 'S'], [147, 'E']],
        'Paracel Islands': [[16, 'N'], [112, 'E']],
        'Paraguay': [[23, 'S'], [58, 'W']],
        'Peru': [[10, 'S'], [76, 'W']],
        'Philippines': [[13, 'N'], [122, 'E']],
        'Pitcairn Islands': [[25, 'S'], [130, 'W']],
        'Poland': [[52, 'N'], [20, 'E']],
        'Portugal': [[39, 'N'], [8, 'W']],
        'Puerto Rico': [[18, 'N'], [66, 'W']],
        'Qatar': [[25, 'N'], [51, 'E']],
        'Republic of the Congo': [[1, 'S'], [15, 'E']],
        'Romania': [[46, 'N'], [25, 'E']],
        'Russia': [[60, 'N'], [100, 'E']],
        'Rwanda': [[2, 'S'], [30, 'E']],
        'Saint Barthelemy': [[17, 'N'], [62, 'W']],
        'Saint Helena': [[15, 'S'], [5, 'W']],
        'Saint Kitts and Nevis': [[17, 'N'], [62, 'W']],
        'Saint Lucia': [[13, 'N'], [60, 'W']],
        'Saint Martin': [[18, 'N'], [63, 'W']],
        'Saint Pierre and Miquelon': [[46, 'N'], [56, 'W']],
        'Saint Vincent and the Grenadines': [[13, 'N'], [61, 'W']],
        'Samoa': [[13, 'S'], [172, 'W']],
        'San Marino': [[43, 'N'], [12, 'E']],
        'Sao Tome and Principe': [[1, 'N'], [7, 'E']],
        'Saudi Arabia': [[25, 'N'], [45, 'E']],
        'Senegal': [[14, 'N'], [14, 'W']],
        'Serbia': [[44, 'N'], [21, 'E']],
        'Seychelles': [[4, 'S'], [55, 'E']],
        'Sierra Leone': [[8, 'N'], [11, 'W']],
        'Singapore': [[1, 'N'], [103, 'E']],
        'Slovakia': [[48, 'N'], [19, 'E']],
        'Slovenia': [[46, 'N'], [14, 'E']],
        'Solomon Islands': [[8, 'S'], [159, 'E']],
        'Somalia': [[10, 'N'], [49, 'E']],
        'South Africa': [[29, 'S'], [24, 'E']],
        'South Georgia and the South Sandwich Islands': [[54, 'S'], [37, 'W']],
        'South Korea': [[37, 'N'], [127, 'E']],
        'Spain': [[40, 'N'], [4, 'W']],
        'Spratly Islands': [[8, 'N'], [111, 'E']],
        'Sri Lanka': [[7, 'N'], [81, 'E']],
        'Sudan': [[15, 'N'], [30, 'E']],
        'Suriname': [[4, 'N'], [56, 'W']],
        'Svalbard': [[78, 'N'], [20, 'E']],
        'Swaziland': [[26, 'S'], [31, 'E']],
        'Sweden': [[62, 'N'], [15, 'E']],
        'Switzerland': [[47, 'N'], [8, 'E']],
        'Syria': [[35, 'N'], [38, 'E']],
        'Taiwan': [[23, 'N'], [121, 'E']],
        'Tajikistan': [[39, 'N'], [71, 'E']],
        'Tanzania': [[6, 'S'], [35, 'E']],
        'Thailand': [[15, 'N'], [100, 'E']],
        'The Gambia': [[13, 'N'], [16, 'W']],
        'Timor-Leste': [[8, 'S'], [125, 'E']],
        'Togo': [[8, 'N'], [1, 'E']],
        'Tokelau': [[9, 'S'], [172, 'W']],
        'Tonga': [[20, 'S'], [175, 'W']],
        'Trinidad and Tobago': [[11, 'N'], [61, 'W']],
        'Tunisia': [[34, 'N'], [9, 'E']],
        'Turkey': [[39, 'N'], [35, 'E']],
        'Turkmenistan': [[40, 'N'], [60, 'E']],
        'Turks and Caicos Islands': [[21, 'N'], [71, 'W']],
        'Tuvalu': [[8, 'S'], [178, 'E']],
        'Uganda': [[1, 'N'], [32, 'E']],
        'Ukraine': [[49, 'N'], [32, 'E']],
        'United Arab Emirates': [[24, 'N'], [54, 'E']],
        'United Kingdom': [[54, 'N'], [2, 'W']],
        'United States': [[38, 'N'], [97, 'W']],
        'Uruguay': [[33, 'S'], [56, 'W']],
        'Uzbekistan': [[41, 'N'], [64, 'E']],
        'Vanuatu': [[16, 'S'], [167, 'E']],
        'Vatican City': [[41, 'N'], [12, 'E']],
        'Venezuela': [[8, 'N'], [66, 'W']],
        'Vietnam': [[16, 'N'], [107, 'E']],
        'Virgin Islands': [[18, 'N'], [64, 'W']],
        'Wake Island': [[19, 'N'], [166, 'E']],
        'Wallis and Futuna': [[13, 'S'], [176, 'W']],
        'West Bank': [[32, 'N'], [35, 'E']],
        'Western Sahara': [[24, 'N'], [13, 'W']],
        'Yemen': [[15, 'N'], [48, 'E']],
        'Zambia': [[15, 'S'], [30, 'E']],
        'Zimbabwe': [[20, 'S'], [30, 'E']]
        }
    g = Graph()
    g.add_edges(edges)
    g.add_vertices(gps_coordinates)
    g.gps_coordinates = gps_coordinates
    g.name("World Map")
    return g



# sage_setup: distribution = sagemath-graphs
# sage.doctest: needs sage.rings.finite_rings
r"""
Block designs

A *block design* is a set together with a family of subsets (repeated subsets
are allowed) whose members are chosen to satisfy some set of properties that are
deemed useful for a particular application. See :wikipedia:`Block_design`.

REFERENCES:

- Block design from wikipedia: :wikipedia:`Block_design`

- What is a block design?,
  http://designtheory.org/library/extrep/extrep-1.1-html/node4.html (in 'The
  External Representation of Block Designs' by Peter J. Cameron, Peter
  Dobcsanyi, John P. Morgan, Leonard H. Soicher)

.. [Hu57] Daniel R. Hughes, "A class of non-Desarguesian projective planes",
   The Canadian Journal of Mathematics (1957), http://cms.math.ca/cjm/v9/p378

.. [We07] Charles Weibel, "Survey of Non-Desarguesian planes" (2007), notices of
   the AMS, vol. 54 num. 10, pages 1294--1303

AUTHORS:

- Quentin Honoré (2015): construction of Hughes plane :issue:`18527`

- Vincent Delecroix (2014): rewrite the part on projective planes :issue:`16281`

- Peter Dobcsanyi and David Joyner (2007-2008)

  This is a significantly modified form of the module block_design.py (version
  0.6) written by Peter Dobcsanyi peter@designtheory.org. Thanks go to Robert
  Miller for lots of good design suggestions.

.. TODO::

    Implement more finite non-Desarguesian plane as in [We07]_ and
    :wikipedia:`Non-Desarguesian_plane`.

Functions and methods
---------------------
"""

# ****************************************************************************
#       Copyright (C) 2007 Peter Dobcsanyi <peter@designtheory.org>
#       Copyright (C) 2007 David Joyner <wdjoyner@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************
from sage.arith.misc import binomial, integer_floor, is_prime_power
from sage.categories.sets_cat import EmptySetError
from sage.misc.lazy_import import lazy_import
from sage.misc.unknown import Unknown
from sage.rings.integer import Integer
from sage.rings.integer_ring import ZZ

from .incidence_structures import IncidenceStructure

lazy_import('sage.libs.gap.libgap', 'libgap')
lazy_import('sage.matrix.matrix_space', 'MatrixSpace')
lazy_import('sage.modules.free_module', 'VectorSpace')
lazy_import('sage.rings.finite_rings.finite_field_constructor', 'FiniteField')


BlockDesign = IncidenceStructure

#  utility functions  -----------------------------------------------------


def tdesign_params(t, v, k, L):
    """
    Return the design's parameters: `(t, v, b, r , k, L)`. Note that `t` must be
    given.

    EXAMPLES::

        sage: BD = BlockDesign(7, [[0,1,2],[0,3,4],[0,5,6],[1,3,5],[1,4,6],[2,3,6],[2,4,5]])
        sage: from sage.combinat.designs.block_design import tdesign_params
        sage: tdesign_params(2,7,3,1)
        (2, 7, 7, 3, 3, 1)
    """
    x = binomial(v, t)
    y = binomial(k, t)
    b = divmod(L * x, y)[0]
    x = binomial(v-1, t-1)
    y = binomial(k-1, t-1)
    r = integer_floor(L * x/y)
    return (t, v, b, r, k, L)


def are_hyperplanes_in_projective_geometry_parameters(v, k, lmbda, return_parameters=False):
    r"""
    Return ``True`` if the parameters ``(v,k,lmbda)`` are the one of hyperplanes in
    a (finite Desarguesian) projective space.

    In other words, test whether there exists a prime power ``q`` and an integer
    ``d`` greater than two such that:

    - `v = (q^{d+1}-1)/(q-1) = q^d + q^{d-1} + ... + 1`
    - `k = (q^d - 1)/(q-1) = q^{d-1} + q^{d-2} + ... + 1`
    - `lmbda = (q^{d-1}-1)/(q-1) = q^{d-2} + q^{d-3} + ... + 1`

    If it exists, such a pair ``(q,d)`` is unique.

    INPUT:

    - ``v``, ``k``, ``lmbda`` -- integers

    OUTPUT:

    - a boolean or, if ``return_parameters`` is set to ``True`` a pair
      ``(True, (q,d))`` or ``(False, (None,None))``.

    EXAMPLES::

        sage: from sage.combinat.designs.block_design import are_hyperplanes_in_projective_geometry_parameters
        sage: are_hyperplanes_in_projective_geometry_parameters(40, 13, 4)
        True
        sage: are_hyperplanes_in_projective_geometry_parameters(40, 13, 4,
        ....:                                                   return_parameters=True)
        (True, (3, 3))
        sage: PG = designs.ProjectiveGeometryDesign(3, 2, GF(3))                        # needs sage.combinat
        sage: PG.is_t_design(return_parameters=True)                                    # needs sage.combinat
        (True, (2, 40, 13, 4))

        sage: are_hyperplanes_in_projective_geometry_parameters(15, 3, 1)
        False
        sage: are_hyperplanes_in_projective_geometry_parameters(15, 3, 1,
        ....:                                                   return_parameters=True)
        (False, (None, None))

    TESTS::

        sage: sgp = lambda q,d: ((q**(d+1)-1)//(q-1), (q**d-1)//(q-1), (q**(d-1)-1)//(q-1))
        sage: for q in [3,4,5,7,8,9,11]:
        ....:     for d in [2,3,4,5]:
        ....:         v,k,l = sgp(q,d)
        ....:         assert are_hyperplanes_in_projective_geometry_parameters(v,k,l,True) == (True, (q,d))
        ....:         assert are_hyperplanes_in_projective_geometry_parameters(v+1,k,l) is False
        ....:         assert are_hyperplanes_in_projective_geometry_parameters(v-1,k,l) is False
        ....:         assert are_hyperplanes_in_projective_geometry_parameters(v,k+1,l) is False
        ....:         assert are_hyperplanes_in_projective_geometry_parameters(v,k-1,l) is False
        ....:         assert are_hyperplanes_in_projective_geometry_parameters(v,k,l+1) is False
        ....:         assert are_hyperplanes_in_projective_geometry_parameters(v,k,l-1) is False
    """
    import sage.arith.all as arith

    q1 = Integer(v - k)
    q2 = Integer(k - lmbda)

    if (lmbda <= 0 or q1 < 4 or q2 < 2 or
        not q1.is_prime_power() or
        not q2.is_prime_power()):
        return (False,(None,None)) if return_parameters else False

    p1,e1 = q1.factor()[0]
    p2,e2 = q2.factor()[0]

    k = arith.gcd(e1,e2)
    d = e1//k
    q = p1**k
    if e2//k != d-1 or lmbda != (q**(d-1)-1)//(q-1):
        return (False,(None,None)) if return_parameters else False

    return (True, (q,d)) if return_parameters else True


def ProjectiveGeometryDesign(n, d, F, algorithm=None, point_coordinates=True, check=True):
    r"""
    Return a projective geometry design.

    The projective geometry design `PG_d(n,q)` has for points the lines of
    `\GF{q}^{n+1}`, and for blocks the `(d+1)`-dimensional subspaces of
    `\GF{q}^{n+1}`, each of which contains `\frac {|\GF{q}|^{d+1}-1} {|\GF{q}|-1}` lines.
    It is a `2`-design with parameters

    .. MATH::

        v = \binom{n+1}{1}_q,\ k = \binom{d+1}{1}_q,\ \lambda =
        \binom{n-1}{d-1}_q

    where the `q`-binomial coefficient `\binom{m}{r}_q` is defined by

    .. MATH::

        \binom{m}{r}_q = \frac{(q^m - 1)(q^{m-1} - 1) \cdots (q^{m-r+1}-1)}
              {(q^r-1)(q^{r-1}-1)\cdots (q-1)}

    .. SEEALSO::

        :func:`AffineGeometryDesign`

    INPUT:

    - ``n`` -- the projective dimension

    - ``d`` -- the dimension of the subspaces which make up the blocks

    - ``F`` -- a finite field or a prime power

    - ``algorithm`` -- set to ``None`` by default, which results in using Sage's
      own implementation. In order to use GAP's implementation instead (i.e. its
      ``PGPointFlatBlockDesign`` function) set ``algorithm="gap"``. Note that
      GAP's "design" package must be available in this case, and that it can be
      installed with the ``gap_packages`` spkg.

    - ``point_coordinates`` -- ``True`` by default. Ignored and assumed to be ``False`` if
      ``algorithm="gap"``. If ``True``, the ground set is indexed by coordinates
      in `\GF{q}^{n+1}`.  Otherwise the ground set is indexed by integers.

    - ``check`` -- boolean (default: ``True``); whether to check the output

    EXAMPLES:

    The set of `d`-dimensional subspaces in a `n`-dimensional projective space
    forms `2`-designs (or balanced incomplete block designs)::

        sage: PG = designs.ProjectiveGeometryDesign(4, 2, GF(2)); PG                    # needs sage.combinat
        Incidence structure with 31 points and 155 blocks
        sage: PG.is_t_design(return_parameters=True)                                    # needs sage.combinat
        (True, (2, 31, 7, 7))

        sage: PG = designs.ProjectiveGeometryDesign(3, 1, GF(4))                        # needs sage.combinat
        sage: PG.is_t_design(return_parameters=True)                                    # needs sage.combinat
        (True, (2, 85, 5, 1))

    Check with ``F`` being a prime power::

        sage: PG = designs.ProjectiveGeometryDesign(3, 2, 4); PG
        Incidence structure with 85 points and 85 blocks

    Use coordinates::

        sage: PG = designs.ProjectiveGeometryDesign(2, 1, GF(3))
        sage: PG.blocks()[0]
        [(1, 0, 0), (1, 0, 1), (1, 0, 2), (0, 0, 1)]

    Use indexing by integers::

        sage: PG = designs.ProjectiveGeometryDesign(2, 1, GF(3), point_coordinates=0)
        sage: PG.blocks()[0]
        [0, 1, 2, 12]

    Check that the constructor using gap also works::

        sage: BD = designs.ProjectiveGeometryDesign(2, 1, GF(2), algorithm='gap')  # optional - gap_package_design
        sage: BD.is_t_design(return_parameters=True)                               # optional - gap_package_design
        (True, (2, 7, 3, 1))
    """
    try:
        q = int(F)
    except TypeError:
        q = F.cardinality()
    else:
        from sage.rings.finite_rings.finite_field_constructor import GF
        F = GF(q)

    if algorithm is None:
        from sage.matrix.echelon_matrix import reduced_echelon_matrix_iterator

        points = {p:i for i,p in enumerate(reduced_echelon_matrix_iterator(F,1,n+1,copy=True,set_immutable=True))}
        blocks = []
        for m1 in reduced_echelon_matrix_iterator(F,d+1,n+1,copy=False):
            b = []
            for m2 in reduced_echelon_matrix_iterator(F,1,d+1,copy=False):
                m = m2*m1
                m.echelonize()
                m.set_immutable()
                b.append(points[m])
            blocks.append(b)
        B = BlockDesign(len(points), blocks, name='ProjectiveGeometryDesign', check=check)
        if point_coordinates:
            B.relabel({i:p[0] for p,i in points.items()})

    elif algorithm == "gap":   # Requires GAP's Design
        libgap.load_package("design")
        D = libgap.PGPointFlatBlockDesign(n, F.order(), d)
        v = D['v'].sage()
        gblcks = D['blocks'].sage()
        gB = []
        for b in gblcks:
            gB.append([x - 1 for x in b])
        B = BlockDesign(v, gB, name='ProjectiveGeometryDesign', check=check)

    if check:
        from sage.combinat.q_analogues import q_binomial
        q = F.cardinality()
        if not B.is_t_design(t=2, v=q_binomial(n+1,1,q),
                                  k=q_binomial(d+1,1,q),
                                  l=q_binomial(n-1, d-1, q)):
            raise RuntimeError("error in ProjectiveGeometryDesign "
                    "construction. Please e-mail sage-devel@googlegroups.com")
    return B


def DesarguesianProjectivePlaneDesign(n, point_coordinates=True, check=True):
    r"""
    Return the Desarguesian projective plane of order ``n`` as a 2-design.

    The Desarguesian projective plane of order `n` can also be defined as the
    projective plane over a field of order `n`. For more information, have a
    look at :wikipedia:`Projective_plane`.

    INPUT:

    - ``n`` -- integer which must be a power of a prime number

    - ``point_coordinates`` -- boolean (default: ``True``); whether to label the
      points with their homogeneous coordinates (default) or with integers

    - ``check`` -- boolean (default: ``True``); whether to check that output is
      correct before returning it. As this is expected to be useless, you may
      want to disable it whenever you want speed.

    .. SEEALSO::

        :func:`ProjectiveGeometryDesign`

    EXAMPLES::

        sage: designs.DesarguesianProjectivePlaneDesign(2)
        (7,3,1)-Balanced Incomplete Block Design
        sage: designs.DesarguesianProjectivePlaneDesign(3)
        (13,4,1)-Balanced Incomplete Block Design
        sage: designs.DesarguesianProjectivePlaneDesign(4)
        (21,5,1)-Balanced Incomplete Block Design
        sage: designs.DesarguesianProjectivePlaneDesign(5)
        (31,6,1)-Balanced Incomplete Block Design
        sage: designs.DesarguesianProjectivePlaneDesign(6)
        Traceback (most recent call last):
        ...
        ValueError: the order of a finite field must be a prime power
    """
    K = FiniteField(n, 'a')
    n2 = n**2
    relabel = {x: i for i, x in enumerate(K)}
    Kiter = relabel
    # it is much faster to iterate through a dict than through
    # the finite field K

    # we decompose the (equivalence class) of points [x:y:z] of the projective
    # plane into an affine plane, an affine line and a point. At the same time,
    # we relabel the points with the integers from 0 to n^2 + n as follows:
    # - the affine plane is the set of points [x:y:1] (i.e. the third coordinate
    #   is nonzero) and gets relabeled from 0 to n^2-1
    affine_plane = lambda x,y: relabel[x] + n * relabel[y]

    # - the affine line is the set of points [x:1:0] (i.e. the third coordinate is
    #   zero but not the second one) and gets relabeled from n^2 to n^2 + n - 1
    line_infinity = lambda x: n2 + relabel[x]

    # - the point is [1:0:0] and gets relabeled n^2 + n
    point_infinity = n2 + n

    blcks = []

    # the n^2 lines of the form "x = sy + az"
    for s in Kiter:
        for a in Kiter:
            # points in the affine plane
            blcks.append([affine_plane(s*y+a, y) for y in Kiter])
            # point at infinity
            blcks[-1].append(line_infinity(s))

    # the n horizontals of the form "y = az"
    for a in Kiter:
        # points in the affine plane
        blcks.append([affine_plane(x,a) for x in Kiter])
        # point at infinity
        blcks[-1].append(point_infinity)

    # the line at infinity "z = 0"
    blcks.append(range(n2,n2+n+1))
    if check:
        from .designs_pyx import is_projective_plane
        if not is_projective_plane(blcks):
            raise RuntimeError('There is a problem in the function DesarguesianProjectivePlane')
    from .bibd import BalancedIncompleteBlockDesign
    B = BalancedIncompleteBlockDesign(n2+n+1, blcks, check=check)

    if point_coordinates:
        zero = K.zero()
        one = K.one()
        d = {affine_plane(x,y): (x,y,one)
             for x in Kiter
             for y in Kiter}
        d.update({line_infinity(x): (x,one,zero)
                  for x in Kiter})
        d[n2+n] = (one,zero,zero)
        B.relabel(d)

    return B


def q3_minus_one_matrix(K):
    r"""
    Return a companion matrix in `GL(3, K)` whose multiplicative order is `q^3 - 1`.

    This function is used in :func:`HughesPlane`.

    EXAMPLES::

        sage: from sage.combinat.designs.block_design import q3_minus_one_matrix
        sage: m = q3_minus_one_matrix(GF(3))
        sage: m.multiplicative_order() == 3**3 - 1
        True

        sage: m = q3_minus_one_matrix(GF(4, 'a'))
        sage: m.multiplicative_order() == 4**3 - 1
        True

        sage: m = q3_minus_one_matrix(GF(5))
        sage: m.multiplicative_order() == 5**3 - 1
        True

        sage: m = q3_minus_one_matrix(GF(9, 'a'))
        sage: m.multiplicative_order() == 9**3 - 1
        True
    """
    q = K.cardinality()
    M = MatrixSpace(K, 3)

    if q.is_prime():
        from sage.rings.finite_rings.conway_polynomials import conway_polynomial
        try:
            a,b,c,_ = conway_polynomial(q, 3)
        except RuntimeError:  # the polynomial is not in the database
            pass
        else:
            return M([0,0,-a,1,0,-b,0,1,-c])

    m = M()
    m[1,0] = m[2,1] = K.one()
    while True:
        m[0,2] = K._random_nonzero_element()
        m[1,2] = K.random_element()
        m[2,2] = K.random_element()
        if m.multiplicative_order() == q**3 - 1:
            return m


def normalize_hughes_plane_point(p, q):
    r"""
    Return the normalized form of point ``p`` as a 3-tuple.

    In the Hughes projective plane over the finite field `K`, all triples `(xk,
    yk, zk)` with `k \in K` represent the same point (where the multiplication
    is over the nearfield built from `K`). This function chooses a canonical
    representative among them.

    This function is used in :func:`HughesPlane`.

    INPUT:

    - ``p`` -- point with the coordinates `(x,y,z)` (a list, a vector, a tuple...)

    - ``q`` -- cardinality of the underlying finite field

    EXAMPLES::

        sage: from sage.combinat.designs.block_design import normalize_hughes_plane_point
        sage: K = FiniteField(9,'x')
        sage: x = K.gen()
        sage: normalize_hughes_plane_point((x, x + 1, x), 9)
        (1, x, 1)
        sage: normalize_hughes_plane_point(vector((x,x,x)), 9)
        (1, 1, 1)
        sage: zero = K.zero()
        sage: normalize_hughes_plane_point((2*x + 2, zero, zero), 9)
        (1, 0, 0)
        sage: one = K.one()
        sage: normalize_hughes_plane_point((2*x, one, zero), 9)
        (2*x, 1, 0)
    """
    for i in [2,1,0]:
        if p[i].is_one():
            return tuple(p)
        elif not p[i].is_zero():
            k = ~p[i]
            if k.is_square():
                return (p[0] * k,p[1] * k,p[2] * k)
            else:
                return ((p[0] * k)**q,(p[1]*k)**q,(p[2]*k)**q)


def HughesPlane(q2, check=True):
    r"""
    Return the Hughes projective plane of order ``q2``.

    Let `q` be an odd prime, the Hughes plane of order `q^2` is a finite
    projective plane of order `q^2` introduced by D. Hughes in [Hu57]_. Its
    construction is as follows.

    Let `K = GF(q^2)` be a finite field with `q^2` elements and `F = GF(q)
    \subset K` be its unique subfield with `q` elements. We define a twisted
    multiplication on `K` as

    .. MATH::

        x \circ y =
        \begin{cases}
        x\ y & \text{if y is a square in K}\\
        x^q\ y & \text{otherwise}
        \end{cases}

    The points of the Hughes plane are the triples `(x, y, z)` of points in `K^3
    \backslash \{0,0,0\}` up to the equivalence relation `(x,y,z) \sim (x \circ
    k, y \circ k, z \circ k)` where `k \in K`.

    For `a = 1` or `a \in (K \backslash F)` we define a block `L(a)` as the set of
    triples `(x,y,z)` so that `x + a \circ y + z = 0`. The rest of the blocks
    are obtained by letting act the group `GL(3, F)` by its standard action.

    For more information, see :wikipedia:`Hughes_plane` and [We07].

    .. SEEALSO::

        :func:`DesarguesianProjectivePlaneDesign` to build the Desarguesian
        projective planes

    INPUT:

    - ``q2`` -- an even power of an odd prime number

    - ``check`` -- boolean (default: ``True``); whether to check that output is
      correct before returning it. As this is expected to be useless, you may
      want to disable it whenever you want speed.

    EXAMPLES::

        sage: H = designs.HughesPlane(9); H
        (91,10,1)-Balanced Incomplete Block Design

    We prove in the following computations that the Desarguesian plane ``H`` is
    not Desarguesian. Let us consider the two triangles `(0,1,10)` and `(57, 70,
    59)`. We show that the intersection points `D_{0,1} \cap D_{57,70}`,
    `D_{1,10} \cap D_{70,59}` and `D_{10,0} \cap D_{59,57}` are on the same line
    while `D_{0,70}`, `D_{1,59}` and `D_{10,57}` are not concurrent::

        sage: blocks = H.blocks()
        sage: line = lambda p,q: next(b for b in blocks if p in b and q in b)

        sage: b_0_1 = line(0, 1)
        sage: b_1_10 = line(1, 10)
        sage: b_10_0 = line(10, 0)
        sage: b_57_70 = line(57, 70)
        sage: b_70_59 = line(70, 59)
        sage: b_59_57 = line(59, 57)

        sage: set(b_0_1).intersection(b_57_70)
        {2}
        sage: set(b_1_10).intersection(b_70_59)
        {73}
        sage: set(b_10_0).intersection(b_59_57)
        {60}

        sage: line(2, 73) == line(73, 60)
        True

        sage: b_0_57 = line(0, 57)
        sage: b_1_70 = line(1, 70)
        sage: b_10_59 = line(10, 59)

        sage: p = set(b_0_57).intersection(b_1_70)
        sage: q = set(b_1_70).intersection(b_10_59)
        sage: p == q
        False

    TESTS:

    Some wrong input::

        sage: designs.HughesPlane(5)
        Traceback (most recent call last):
        ...
        EmptySetError: No Hughes plane of non-square order exists.

        sage: designs.HughesPlane(16)
        Traceback (most recent call last):
        ...
        EmptySetError: No Hughes plane of even order exists.

    Check that it works for non-prime `q`::

        sage: designs.HughesPlane(3**4)    # not tested - 10 secs
        (6643,82,1)-Balanced Incomplete Block Design
    """
    if not q2.is_square():
        raise EmptySetError("No Hughes plane of non-square order exists.")
    if q2 % 2 == 0:
        raise EmptySetError("No Hughes plane of even order exists.")
    q = q2.sqrt()
    K = FiniteField(q2, prefix='x')
    F = FiniteField(q, prefix='y')
    A = q3_minus_one_matrix(F)
    A = A.change_ring(K)
    m = K.list()
    V = VectorSpace(K, 3)
    zero = K.zero()
    one = K.one()
    points = [(x, y, one) for x in m for y in m] + \
             [(x, one, zero) for x in m] + \
             [(one, zero, zero)]
    relabel = {tuple(p):i for i,p in enumerate(points)}
    blcks = []
    for a in m:
        if a not in F or a == 1:
            # build L(a)
            aa = ~a
            l = []
            l.append(V((-a, one, zero)))
            for x in m:
                y = - aa * (x+one)
                if not y.is_square():
                    y *= aa**(q-1)
                l.append(V((x, y, one)))
            # compute the orbit of L(a)
            blcks.append([relabel[normalize_hughes_plane_point(p,q)] for p in l])
            for i in range(q2 + q):
                l = [A*j for j in l]
                blcks.append([relabel[normalize_hughes_plane_point(p,q)] for p in l])
    from .bibd import BalancedIncompleteBlockDesign
    return BalancedIncompleteBlockDesign(q2**2+q2+1, blcks, check=check)


def projective_plane_to_OA(pplane, pt=None, check=True):
    r"""
    Return the orthogonal array built from the projective plane ``pplane``.

    The orthogonal array `OA(n+1,n,2)` is obtained from the projective plane
    ``pplane`` by removing the point ``pt`` and the `n+1` lines that pass
    through it`. These `n+1` lines form the `n+1` groups while the remaining
    `n^2+n` lines form the transversals.

    INPUT:

    - ``pplane`` -- a projective plane as a 2-design

    - ``pt`` -- a point in the projective plane ``pplane``; if it is not provided,
      then it is set to `n^2 + n`

    - ``check`` -- boolean (default: ``True``); whether to check that output is
      correct before returning it. As this is expected to be useless, you may
      want to disable it whenever you want speed.

    EXAMPLES::

        sage: from sage.combinat.designs.block_design import projective_plane_to_OA
        sage: p2 = designs.DesarguesianProjectivePlaneDesign(2, point_coordinates=False)
        sage: projective_plane_to_OA(p2)
        [[0, 0, 0], [0, 1, 1], [1, 0, 1], [1, 1, 0]]
        sage: p3 = designs.DesarguesianProjectivePlaneDesign(3, point_coordinates=False)
        sage: projective_plane_to_OA(p3)
        [[0, 0, 0, 0],
         [0, 1, 2, 1],
         [0, 2, 1, 2],
         [1, 0, 2, 2],
         [1, 1, 1, 0],
         [1, 2, 0, 1],
         [2, 0, 1, 1],
         [2, 1, 0, 2],
         [2, 2, 2, 0]]

        sage: pp = designs.DesarguesianProjectivePlaneDesign(16, point_coordinates=False)
        sage: _ = projective_plane_to_OA(pp, pt=0)
        sage: _ = projective_plane_to_OA(pp, pt=3)
        sage: _ = projective_plane_to_OA(pp, pt=7)
    """
    from .bibd import _relabel_bibd
    pplane = pplane.blocks()
    n = len(pplane[0]) - 1

    if pt is None:
        pt = n**2+n

    assert len(pplane) == n**2+n+1, "pplane is not a projective plane"
    assert all(len(B) == n+1 for B in pplane), "pplane is not a projective plane"

    pplane = _relabel_bibd(pplane,n**2+n+1,p=n**2+n)
    OA = [[x % n for x in sorted(X)] for X in pplane if n**2+n not in X]

    assert len(OA) == n**2, "pplane is not a projective plane"

    if check:
        from .designs_pyx import is_orthogonal_array
        is_orthogonal_array(OA,n+1,n,2)

    return OA


def projective_plane(n, check=True, existence=False):
    r"""
    Return a projective plane of order ``n`` as a 2-design.

    A finite projective plane is a 2-design with `n^2+n+1` lines (or blocks) and
    `n^2+n+1` points. For more information on finite projective planes, see the
    :wikipedia:`Projective_plane#Finite_projective_planes`.

    If no construction is possible, then the function raises a :exc:`EmptySetError`,
    whereas if no construction is available, the function raises a
    :exc:`NotImplementedError`.

    INPUT:

    - ``n`` -- the finite projective plane's order

    EXAMPLES::

        sage: # needs sage.schemes
        sage: designs.projective_plane(2)
        (7,3,1)-Balanced Incomplete Block Design
        sage: designs.projective_plane(3)
        (13,4,1)-Balanced Incomplete Block Design
        sage: designs.projective_plane(4)
        (21,5,1)-Balanced Incomplete Block Design
        sage: designs.projective_plane(5)
        (31,6,1)-Balanced Incomplete Block Design
        sage: designs.projective_plane(6)
        Traceback (most recent call last):
        ...
        EmptySetError: By the Bruck-Ryser theorem, no projective plane of order 6 exists.
        sage: designs.projective_plane(10)
        Traceback (most recent call last):
        ...
        EmptySetError: No projective plane of order 10 exists by
        C. Lam, L. Thiel and S. Swiercz "The nonexistence of finite
        projective planes of order 10" (1989), Canad. J. Math.
        sage: designs.projective_plane(12)
        Traceback (most recent call last):
        ...
        NotImplementedError: If such a projective plane exists,
        we do not know how to build it.
        sage: designs.projective_plane(14)
        Traceback (most recent call last):
        ...
        EmptySetError: By the Bruck-Ryser theorem, no projective plane of order 14 exists.

    TESTS::

        sage: # needs sage.schemes
        sage: designs.projective_plane(2197, existence=True)
        True
        sage: designs.projective_plane(6, existence=True)
        False
        sage: designs.projective_plane(10, existence=True)
        False
        sage: designs.projective_plane(12, existence=True)
        Unknown
    """
    from sage.combinat.designs.bibd import BruckRyserChowla_check

    if n <= 1:
        if existence:
            return False
        raise EmptySetError("There is no projective plane of order <= 1")

    if n == 10:
        if existence:
            return False
        ref = ("C. Lam, L. Thiel and S. Swiercz \"The nonexistence of finite "
               "projective planes of order 10\" (1989), Canad. J. Math.")
        raise EmptySetError("No projective plane of order 10 exists by %s" % ref)

    if BruckRyserChowla_check(n*n+n+1, n+1, 1) is False:
        if existence:
            return False
        raise EmptySetError("By the Bruck-Ryser theorem, no projective"
                         " plane of order {} exists.".format(n))

    if not is_prime_power(n):
        if existence:
            return Unknown
        raise NotImplementedError("If such a projective plane exists, we do "
                                  "not know how to build it.")

    if existence:
        return True
    else:
        return DesarguesianProjectivePlaneDesign(n, point_coordinates=False, check=check)


def AffineGeometryDesign(n, d, F, point_coordinates=True, check=True):
    r"""
    Return an affine geometry design.

    The affine geometry design `AG_d(n,q)` is the 2-design whose blocks are the
    `d`-vector subspaces in `\GF{q}^n`. It has parameters

    .. MATH::

        v = q^n,\ k = q^d,\ \lambda = \binom{n-1}{d-1}_q

    where the `q`-binomial coefficient `\binom{m}{r}_q` is defined by

    .. MATH::

        \binom{m}{r}_q = \frac{(q^m - 1)(q^{m-1} - 1) \cdots (q^{m-r+1}-1)}
              {(q^r-1)(q^{r-1}-1)\cdots (q-1)}

    .. SEEALSO::

        :func:`ProjectiveGeometryDesign`

    INPUT:

    - ``n`` -- integer; the Euclidean dimension. The number of points of the
      design is `v=|\GF{q}^n|`

    - ``d`` -- integer; the dimension of the (affine) subspaces of `\GF{q}^n`
      which make up the blocks

    - ``F`` -- a finite field or a prime power

    - ``point_coordinates`` -- boolean (default: ``True``); whether we use
      coordinates in `\GF{q}^n` or plain integers for the points of the design

    - ``check`` -- boolean (default: ``True``); whether to check the output

    EXAMPLES::

        sage: # needs sage.combinat
        sage: BD = designs.AffineGeometryDesign(3, 1, GF(2))
        sage: BD.is_t_design(return_parameters=True)
        (True, (2, 8, 2, 1))
        sage: BD = designs.AffineGeometryDesign(3, 2, GF(4))
        sage: BD.is_t_design(return_parameters=True)
        (True, (2, 64, 16, 5))
        sage: BD = designs.AffineGeometryDesign(4, 2, GF(3))
        sage: BD.is_t_design(return_parameters=True)
        (True, (2, 81, 9, 13))

    With ``F`` an integer instead of a finite field::

        sage: BD = designs.AffineGeometryDesign(3, 2, 4)
        sage: BD.is_t_design(return_parameters=True)
        (True, (2, 64, 16, 5))

    Testing the option ``point_coordinates``::

        sage: designs.AffineGeometryDesign(3, 1, GF(2),
        ....:                              point_coordinates=True).blocks()[0]
        [(0, 0, 0), (0, 0, 1)]
        sage: designs.AffineGeometryDesign(3, 1, GF(2),
        ....:                              point_coordinates=False).blocks()[0]
        [0, 1]
    """
    try:
        q = int(F)
    except TypeError:
        q = F.cardinality()
    else:
        from sage.rings.finite_rings.finite_field_constructor import GF
        F = GF(q)

    n = int(n)
    d = int(d)

    from itertools import islice
    from sage.combinat.q_analogues import q_binomial
    from sage.matrix.echelon_matrix import reduced_echelon_matrix_iterator

    points = {p:i for i,p in enumerate(reduced_echelon_matrix_iterator(F,1,n+1,copy=True,set_immutable=True)) if p[0,0]}

    blocks = []
    l1 = int(q_binomial(n+1, d+1, q) - q_binomial(n, d+1, q))
    l2 = q**d
    for m1 in islice(reduced_echelon_matrix_iterator(F,d+1,n+1,copy=False),
                     int(l1)):
        b = []
        for m2 in islice(reduced_echelon_matrix_iterator(F,1,d+1,copy=False),
                         int(l2)):
            m = m2*m1
            m.echelonize()
            m.set_immutable()
            b.append(points[m])
        blocks.append(b)

    B = BlockDesign(len(points), blocks, name='AffineGeometryDesign', check=check)

    if point_coordinates:
        rd = {i: p[0][1:] for p, i in points.items()}
        for v in rd.values():
            v.set_immutable()
        B.relabel(rd)

    if check:
        if not B.is_t_design(t=2, v=q**n, k=q**d, l=q_binomial(n-1, d-1, q)):
            raise RuntimeError("error in AffineGeometryDesign "
                    "construction. Please e-mail sage-devel@googlegroups.com")
    return B


def CremonaRichmondConfiguration():
    r"""
    Return the Cremona-Richmond configuration.

    The Cremona-Richmond configuration is a set system whose incidence graph
    is equal to the
    :meth:`~sage.graphs.graph_generators.GraphGenerators.TutteCoxeterGraph`. It
    is a generalized quadrangle of parameters `(2,2)`.

    For more information, see the
    :wikipedia:`Cremona-Richmond_configuration`.

    EXAMPLES::

        sage: H = designs.CremonaRichmondConfiguration(); H                             # needs networkx
        Incidence structure with 15 points and 15 blocks
        sage: g = graphs.TutteCoxeterGraph()                                            # needs networkx
        sage: H.incidence_graph().is_isomorphic(g)                                      # needs networkx
        True
    """
    from sage.graphs.generators.smallgraphs import TutteCoxeterGraph
    from sage.combinat.designs.incidence_structures import IncidenceStructure
    g = TutteCoxeterGraph()
    H = IncidenceStructure([g.neighbors(v)
                            for v in g.bipartite_sets()[0]])
    H.relabel()
    return H


def WittDesign(n):
    """
    INPUT:

    - ``n`` -- integer in `9,10,11,12,21,22,23,24`

    Wraps GAP Design's WittDesign. If ``n=24`` then this function returns the
    large Witt design `W_{24}`, the unique (up to isomorphism) `5-(24,8,1)`
    design. If ``n=12`` then this function returns the small Witt design
    `W_{12}`, the unique (up to isomorphism) `5-(12,6,1)` design.  The other
    values of `n` return a block design derived from these.

    .. NOTE::

        Requires GAP's Design package (included in the gap_packages Sage spkg).

    EXAMPLES::

        sage: # optional - gap_package_design
        sage: BD = designs.WittDesign(9)
        sage: BD.is_t_design(return_parameters=True)
        (True, (2, 9, 3, 1))
        sage: BD
        Incidence structure with 9 points and 12 blocks
        sage: print(BD)
        Incidence structure with 9 points and 12 blocks
    """
    libgap.load_package("design")
    B = libgap.WittDesign(n)
    v = B['v'].sage()
    gB = [[x - 1 for x in b] for b in B['blocks'].sage()]
    return BlockDesign(v, gB, name='WittDesign', check=True)


def HadamardDesign(n):
    """
    As described in Section 1, p. 10, in [CvL]_. The input n must have the
    property that there is a Hadamard matrix of order `n+1` (and that a
    construction of that Hadamard matrix has been implemented...).

    EXAMPLES::

        sage: # needs sage.combinat sage.modules
        sage: designs.HadamardDesign(7)
        Incidence structure with 7 points and 7 blocks
        sage: print(designs.HadamardDesign(7))
        Incidence structure with 7 points and 7 blocks

    For example, the Hadamard 2-design with `n = 11` is a design whose parameters are `2-(11, 5, 2)`.
    We verify that `NJ = 5J` for this design. ::

        sage: # needs sage.combinat sage.modules
        sage: D = designs.HadamardDesign(11); N = D.incidence_matrix()
        sage: J = matrix(ZZ, 11, 11, [1]*11*11); N*J
        [5 5 5 5 5 5 5 5 5 5 5]
        [5 5 5 5 5 5 5 5 5 5 5]
        [5 5 5 5 5 5 5 5 5 5 5]
        [5 5 5 5 5 5 5 5 5 5 5]
        [5 5 5 5 5 5 5 5 5 5 5]
        [5 5 5 5 5 5 5 5 5 5 5]
        [5 5 5 5 5 5 5 5 5 5 5]
        [5 5 5 5 5 5 5 5 5 5 5]
        [5 5 5 5 5 5 5 5 5 5 5]
        [5 5 5 5 5 5 5 5 5 5 5]
        [5 5 5 5 5 5 5 5 5 5 5]

    REFERENCES:

    - [CvL] P. Cameron, J. H. van Lint, Designs, graphs, codes and
      their links, London Math. Soc., 1991.
    """
    from sage.combinat.matrices.hadamard_matrix import hadamard_matrix
    from sage.matrix.constructor import matrix
    H = hadamard_matrix(n + 1)  # assumed to be normalised.
    H1 = H.matrix_from_columns(range(1,n+1))
    H2 = H1.matrix_from_rows(range(1,n+1))
    J = matrix(ZZ,n,n,[1]*n*n)
    MS = J.parent()
    A = MS((H2+J)/2)  # convert -1's to 0's; coerce entries to ZZ
    # A is the incidence matrix of the block design
    return IncidenceStructure(incidence_matrix=A, name='HadamardDesign')


def Hadamard3Design(n):
    r"""
    Return the Hadamard 3-design with parameters `3-(n, \frac n 2, \frac n 4 - 1)`.

    This is the unique extension of the Hadamard `2`-design (see
    :meth:`HadamardDesign`).  We implement the description from pp. 12 in
    [CvL]_.

    INPUT:

    - ``n`` -- integer; a multiple of 4 such that `n>4`

    EXAMPLES::

        sage: # needs sage.combinat sage.modules
        sage: designs.Hadamard3Design(12)
        Incidence structure with 12 points and 22 blocks

    We verify that any two blocks of the Hadamard `3`-design `3-(8, 4, 1)`
    design meet in `0` or `2` points. More generally, it is true that any two
    blocks of a Hadamard `3`-design meet in `0` or `\frac{n}{4}` points (for `n
    > 4`).

    ::

        sage: # needs sage.combinat sage.modules
        sage: D = designs.Hadamard3Design(8)
        sage: N = D.incidence_matrix()
        sage: N.transpose()*N
        [4 2 2 2 2 2 2 2 2 2 2 2 2 0]
        [2 4 2 2 2 2 2 2 2 2 2 2 0 2]
        [2 2 4 2 2 2 2 2 2 2 2 0 2 2]
        [2 2 2 4 2 2 2 2 2 2 0 2 2 2]
        [2 2 2 2 4 2 2 2 2 0 2 2 2 2]
        [2 2 2 2 2 4 2 2 0 2 2 2 2 2]
        [2 2 2 2 2 2 4 0 2 2 2 2 2 2]
        [2 2 2 2 2 2 0 4 2 2 2 2 2 2]
        [2 2 2 2 2 0 2 2 4 2 2 2 2 2]
        [2 2 2 2 0 2 2 2 2 4 2 2 2 2]
        [2 2 2 0 2 2 2 2 2 2 4 2 2 2]
        [2 2 0 2 2 2 2 2 2 2 2 4 2 2]
        [2 0 2 2 2 2 2 2 2 2 2 2 4 2]
        [0 2 2 2 2 2 2 2 2 2 2 2 2 4]


    REFERENCES:

    .. [CvL] \P. Cameron, J. H. van Lint, Designs, graphs, codes and
      their links, London Math. Soc., 1991.
    """
    if n == 1 or n == 4:
        raise ValueError("The Hadamard design with n = %s does not extend to a three design." % n)
    from sage.combinat.matrices.hadamard_matrix import hadamard_matrix
    from sage.matrix.constructor import matrix, block_matrix
    H = hadamard_matrix(n)  # assumed to be normalised.
    H1 = H.matrix_from_columns(range(1, n))
    J = matrix(ZZ, n, n-1, [1]*(n-1)*n)
    A1 = (H1 + J) / 2
    A2 = (J - H1) / 2
    A = block_matrix(1, 2, [A1, A2])  # the incidence matrix of the design.
    return IncidenceStructure(incidence_matrix=A, name='HadamardThreeDesign')
